@@ -30,10 +30,9 @@ import shutil
 import tempfile
 import pathlib
 import importlib
+import dill
 
-date = datetime.datetime.now()
-begin_time = str(date)[2:4]+str(date)[5:7]+str(date)[8:10]+"_"+str(date)[11:13]+str(date)[14:16]+str(date)[17:19]
-temp_folder = os.path.join(tempfile.gettempdir(), "gg_"+begin_time)
+forced_structures = ['none', 'n_glycans', 'o_glycans', 'gags']
 
 def config_handler(from_GUI = False, param_file_path = ''):
     '''A function that handles the input of configs through a pipelined
@@ -57,16 +56,22 @@ def config_handler(from_GUI = False, param_file_path = ''):
     functions arguments
         Organized arguments for the various functions used in core module.
     '''
+    
+    date = datetime.datetime.now()
+    begin_time = str(date)[2:4]+str(date)[5:7]+str(date)[8:10]+"_"+str(date)[11:13]+str(date)[14:16]+str(date)[17:19]
+    
     custom_glycans_list = [False, []]
     min_max_monos = [0, 0]
     min_max_hex = [0, 0]
+    min_max_hn = [0, 0]
     min_max_hexnac = [0, 0]
     min_max_xyl = [0, 0]
     min_max_sia = [0, 0]
     min_max_fuc = [0, 0]
     min_max_ac = [0, 0]
     min_max_gc = [0, 0]
-    force_nglycan = False
+    min_max_ua = [0, 0]
+    forced = 'none'
     max_adducts = {}
     adducts_exclusion = []
     max_charges = 0
@@ -74,9 +79,11 @@ def config_handler(from_GUI = False, param_file_path = ''):
     permethylated = False
     reduced = False
     lactonized_ethyl_esterified = False
+    min_max_sulfation = [0, 0]
+    min_max_phosphorylation = [0, 0]
     fast_iso = True
     high_res = False
-    internal_standard = 0.0
+    internal_standard = '0.0'
     imp_exp_library = [False, False]
     exp_lib_name = ''
     library_path = ''
@@ -94,6 +101,7 @@ def config_handler(from_GUI = False, param_file_path = ''):
     close_peaks = [False, 3]
     align_chromatograms = True
     percentage_auc = 0.1
+    min_samples = 0
     max_ppm = 10
     iso_fit_score = 0.9
     curve_fit_score = 0.9
@@ -111,7 +119,8 @@ def config_handler(from_GUI = False, param_file_path = ''):
     samples_list = []
     samples_names = []
     
-    CLI.print_header(False)
+    if not from_GUI:
+        CLI.print_header(False)
     config = configparser.ConfigParser()
     configs = ""
     if from_GUI:
@@ -128,13 +137,9 @@ def config_handler(from_GUI = False, param_file_path = ''):
     save_path = config['running_modes']['working_directory']
     save_path = save_path.strip()
     save_path = save_path.strip("'")
-    save_path = save_path.strip("\"")
-    for i_i, i in enumerate(save_path):
-        if i == "\\":
-            save_path = save_path[:i_i]+"/"+save_path[i_i+1:]
-    if save_path[-1] != "/":
-        save_path = save_path+"/"
-    Path(save_path).mkdir(exist_ok = True, parents = True)
+    save_path = save_path.strip('"')
+    save_path = Path(save_path)
+    save_path.mkdir(exist_ok = True, parents = True)
     
     #multithreading settings
     multithreaded_analysis = config['running_modes'].getboolean('use_multiple_CPU_cores')
@@ -148,10 +153,8 @@ def config_handler(from_GUI = False, param_file_path = ''):
         reanalysis_path = config['running_modes']['file_for_reanalysis']
         reanalysis_path = reanalysis_path.strip()
         reanalysis_path = reanalysis_path.strip("'")
-        reanalysis_path = reanalysis_path.strip("\"")
-        for i_i, i in enumerate(reanalysis_path):
-            if i == "\\":
-                reanalysis_path = reanalysis_path[:i_i]+"/"+reanalysis_path[i_i+1:]
+        reanalysis_path = reanalysis_path.strip('"')
+        reanalysis_path = Path(reanalysis_path)
         
     elif general_mode == 'library':
         only_gen_lib = True
@@ -176,39 +179,37 @@ def config_handler(from_GUI = False, param_file_path = ''):
         library_path = config['library_building_modes']['import_library_path']
         library_path = library_path.strip()
         library_path = library_path.strip("'")
-        library_path = library_path.strip("\"")
-        for i_i, i in enumerate(library_path):
-            if i == "\\":
-                library_path = library_path[:i_i]+"/"+library_path[i_i+1:]
-                
-        shutil.copy(library_path, os.path.join(temp_folder, 'glycans_library.py'))
-        spec = importlib.util.spec_from_file_location("glycans_library", temp_folder+"/glycans_library.py")
-        lib_module = importlib.util.module_from_spec(spec)
-        spec.loader.exec_module(lib_module)
-        try:
-            library_metadata = lib_module.metadata
-        except:
-            library_metadata = []
-        if len(library_metadata) > 0:
-            min_max_monos = library_metadata[0]
-            min_max_hex = library_metadata[1]
-            min_max_hexnac = library_metadata[2]
-            min_max_fuc = library_metadata[3]
-            min_max_sia = library_metadata[4]
-            min_max_ac = library_metadata[5]
-            min_max_gc = library_metadata[6]
-            force_nglycan = library_metadata[7]
-            max_adducts = library_metadata[8]
-            max_charges = library_metadata[9]
-            tag_mass = library_metadata[10]
-            internal_standard = library_metadata[11]
-            permethylated = library_metadata[12]
-            lactonized_ethyl_esterified = library_metadata[13]
-            reduced = library_metadata[14]
-            fast_iso = library_metadata[15]
-            high_res = library_metadata[16]
-            if len(library_metadata) > 18:
-                min_max_xyl = library_metadata[18]
+        library_path = library_path.strip('"')
+        library_path = Path(library_path)
+        
+        with open(library_path, 'rb') as f:
+            library_data = dill.load(f)
+            f.close()
+        full_library = library_data[0]
+        library_metadata = library_data[1]
+        
+        min_max_monos = library_metadata[0]
+        min_max_hex = library_metadata[1]
+        min_max_hexnac = library_metadata[2]
+        min_max_fuc = library_metadata[3]
+        min_max_sia = library_metadata[4]
+        min_max_ac = library_metadata[5]
+        min_max_gc = library_metadata[6]
+        forced = library_metadata[7]
+        max_adducts = library_metadata[8]
+        max_charges = library_metadata[9]
+        tag_mass = library_metadata[10]
+        internal_standard = library_metadata[11]
+        permethylated = library_metadata[12]
+        lactonized_ethyl_esterified = library_metadata[13]
+        reduced = library_metadata[14]
+        fast_iso = library_metadata[15]
+        high_res = library_metadata[16]
+        min_max_xyl = library_metadata[18]
+        min_max_hn = library_metadata[19]
+        min_max_ua = library_metadata[20]
+        min_max_sulfation = library_metadata[21]
+        min_max_phosphorylation = library_metadata[22]
                 
     elif library_mode == 'custom_library':
         custom_glycans_list[0] = True
@@ -216,18 +217,26 @@ def config_handler(from_GUI = False, param_file_path = ''):
         if not reanalysis:
             if (len(config['library_building_modes']['custom_glycans_list'].split("\\")) > 1 or len(config['library_building_modes']['custom_glycans_list'].split("/")) > 1):
                 try:
-                    temp_path_custom_glycans = config['library_building_modes']['custom_glycans_list'].strip("\"").strip("'")
-                    for i_i, i in enumerate(temp_path_custom_glycans):
-                        if i == "\\":
-                            temp_path_custom_glycans = temp_path_custom_glycans[:i_i]+"/"+temp_path_custom_glycans[i_i+1:]
+                    # Adjust custom_glycans path
+                    temp_path_custom_glycans = config['library_building_modes']['custom_glycans_list']
+                    temp_path_custom_glycans = temp_path_custom_glycans.strip()
+                    temp_path_custom_glycans = temp_path_custom_glycans.strip('"')
+                    temp_path_custom_glycans = temp_path_custom_glycans.strip("'")
+                    temp_path_custom_glycans = Path(temp_path_custom_glycans)
+                    
+                    # Extract the list of glycans
                     temp_custom_glycans_list = ""
-                    with open(temp_path_custom_glycans.strip("\""), 'r') as f:
+                    with open(temp_path_custom_glycans, 'r') as f:
                         for i in f:
                             temp_custom_glycans_list += i
-                    f.close()
+                        f.close()
+                    
+                    # Separate the glycans in a list, doesn't matter if they are comma separated or line separated
                     custom_glycans = temp_custom_glycans_list.split(",")
                     if len(custom_glycans) == 1:
                         custom_glycans = temp_custom_glycans_list.split("\n")
+                        
+                    # Sort out the glycans in the list
                     to_remove = []
                     to_add = []
                     if len(custom_glycans) > 1:
@@ -301,10 +310,31 @@ def config_handler(from_GUI = False, param_file_path = ''):
         total_neu5gc = config['library_building_modes']['neu5gc'].split(',')
         min_max_gc = [int(total_neu5gc[0]), int(total_neu5gc[1])]
         
+        total_hn = config['library_building_modes']['hexosamines'].split(',')
+        min_max_hn = [int(total_hn[0]), int(total_hn[1])]
+        
+        total_ua = config['library_building_modes']['uronic_acids'].split(',')
+        min_max_ua = [int(total_ua[0]), int(total_ua[1])]
+        
     imp_exp_library[1] = config['library_building_modes'].getboolean('export_library')
     if imp_exp_library[1]:
         exp_lib_name = config['library_building_modes']['exported_library_name'].strip()
-    force_nglycan = config['common_library_building_settings'].getboolean('force_nglycan')
+    if 'force_nglycan' in config['common_library_building_settings']:
+        if config['common_library_building_settings'].getboolean('force_nglycan'):
+            forced = 'n_glycans'
+        else:
+            forced = 'none'
+    else:
+        if config['common_library_building_settings']['force_class_structure'].lower() in forced_structures:
+            forced = config['common_library_building_settings']['force_class_structure'].lower()
+        else:
+            print("Wrong 'force class structure' chosen.\nYou must choose between 'n_glycan', 'o_glycan'\nor 'gags'.")
+            print("Close the window or press CTRL+C to exit.")
+            try:
+                while True:
+                    time.sleep(3600)
+            except KeyboardInterrupt:
+                os._exit(1)
     max_adducts = General_Functions.form_to_comp(config['common_library_building_settings']['max_adducts'])
     adducts_exc = config['common_library_building_settings']['adducts_exclusion'].split(",")
     if len(adducts_exc) > 0:
@@ -315,13 +345,20 @@ def config_handler(from_GUI = False, param_file_path = ''):
     permethylated = config['common_library_building_settings'].getboolean('permethylated')
     reduced = config['common_library_building_settings'].getboolean('reduced')
     lactonized_ethyl_esterified = config['common_library_building_settings'].getboolean('aminated_ethyl_esterified')
+        
+    total_sulfation = config['common_library_building_settings']['min_max_sulfation_per_glycan'].split(',')
+    min_max_sulfation = [int(total_sulfation[0]), int(total_sulfation[1])]
+        
+    total_phosphorylation = config['common_library_building_settings']['min_max_phosphorylation_per_glycan'].split(',')
+    min_max_phosphorylation = [int(total_phosphorylation[0]), int(total_phosphorylation[1])]
+        
     fast_iso = config['common_library_building_settings'].getboolean('fast_iso')
     high_res = config['common_library_building_settings'].getboolean('high_resolution_isotopic_dist')
     internal_standard = config['common_library_building_settings']['internal_standard_mass']
     if len(internal_standard.strip()) == 0:
-        internal_standard = 0.0
+        internal_standard = '0.0'
     else:
-        internal_standard = float(internal_standard)
+        internal_standard = internal_standard
         
         
     #analysis running mode
@@ -329,14 +366,12 @@ def config_handler(from_GUI = False, param_file_path = ''):
         samples_path = config['running_modes']['samples_directory']
         samples_path = samples_path.strip()
         samples_path = samples_path.strip("'")
-        samples_path = samples_path.strip("\"")
-        for i_i, i in enumerate(samples_path):
-            if i == "\\":
-                samples_path = samples_path[:i_i]+"/"+samples_path[i_i+1:]
-        if samples_path[-1] != "/":
-            samples_path = samples_path+"/"
+        samples_path = samples_path.strip('"')
+        samples_path = Path(samples_path)
+        
         samples_list = Execution_Functions.samples_path_to_list(samples_path)
-        if len(samples_list) == 0 and not from_GUI:                                             
+        
+        if len(samples_list) == 0 and not from_GUI:
             if not os.isatty(0):
                 Execution_Functions.print_sep()
                 print("No sample files to analyze.")
@@ -347,14 +382,15 @@ def config_handler(from_GUI = False, param_file_path = ''):
                 os._exit(1)
         
         samples_names = Execution_Functions.sample_names(samples_list)
-        print("Sample files detected: "+str(len(samples_names)))
-        for i in samples_names:
-            print("--> "+i)
-        Execution_Functions.print_sep()
+        if not from_GUI:
+            print("Sample files detected: "+str(len(samples_names)))
+            for i in samples_names:
+                print("--> "+i)
+            Execution_Functions.print_sep()
             
         ms2_analysis = config['analysis_parameters'].getboolean('analyze_ms2')
         if ms2_analysis:
-            analyze_ms2 = (ms2_analysis, config['analysis_parameters'].getboolean('force_fragments_to_glycans'), config['analysis_parameters'].getboolean('unrestricted_fragments'))
+            analyze_ms2 = [ms2_analysis, config['analysis_parameters'].getboolean('force_fragments_to_glycans'), config['analysis_parameters'].getboolean('unrestricted_fragments')]
             if 'ret_time_tolerance_ms2' in config['analysis_parameters']:
                 rt_tolerance_frag = float(config['analysis_parameters']['ret_time_tolerance_ms2'])
         accuracy_unit = config['analysis_parameters']['accuracy_unit']
@@ -367,15 +403,15 @@ def config_handler(from_GUI = False, param_file_path = ''):
             ret_time_interval[2] = float(config['analysis_parameters']['ret_time_tolerance'])
         custom_ppp = config['analysis_parameters'].getboolean('custom_min_points_per_peak')
         if custom_ppp:
-            min_ppp = (config['analysis_parameters'].getboolean('custom_min_points_per_peak'), int(config['analysis_parameters']['number_points_per_peak']))
+            min_ppp = [config['analysis_parameters'].getboolean('custom_min_points_per_peak'), int(config['analysis_parameters']['number_points_per_peak'])]
         limit_peaks = config['analysis_parameters'].getboolean('limit_peaks_picked')
         if limit_peaks:
-            close_peaks = (config['analysis_parameters'].getboolean('limit_peaks_picked'), int(config['analysis_parameters']['max_number_peaks']))
+            close_peaks = [config['analysis_parameters'].getboolean('limit_peaks_picked'), int(config['analysis_parameters']['max_number_peaks'])]
         if 'noise_levels' in config['analysis_parameters']:
             noise_levels = config['analysis_parameters']['noise_levels'].split(",")
             for i_i, i in enumerate(noise_levels):
                 noise_levels[i_i] = int(i.strip())
-            custom_noise = (config['analysis_parameters'].getboolean('custom_noise_level'), noise_levels)
+            custom_noise = [config['analysis_parameters'].getboolean('custom_noise_level'), noise_levels]
         
     #post-analysis for analysis running mode and reanalysis running mode
     if not only_gen_lib:
@@ -387,11 +423,12 @@ def config_handler(from_GUI = False, param_file_path = ''):
                     reporter_ions = reporter_ions[:i_i]+reporter_ions[i_i+1:]
         align_chromatograms = config['post-analysis/reanalysis'].getboolean('align_chromatograms')
         percentage_auc = float(config['post-analysis/reanalysis']['auc_percentage_threshold'])/100
+        min_samples = int(config['post-analysis/reanalysis']['minimum_samples'])
         ppm_setting = config['post-analysis/reanalysis']['max_ppm_threshold'].split(",")
         if len(ppm_setting) > 1:
-            max_ppm = float(config['post-analysis/reanalysis']['max_ppm_threshold'])
+            max_ppm = [float(ppm_setting[0]), float(ppm_setting[1])]
         else:
-            max_ppm = (float(ppm_setting[0]), float(ppm_setting[1]))
+            max_ppm = float(config['post-analysis/reanalysis']['max_ppm_threshold'])
         iso_fit_score = float(config['post-analysis/reanalysis']['isotopic_fitting_score_threshold'])
         curve_fit_score = float(config['post-analysis/reanalysis']['curve_fitting_score_threshold'])
         s_to_n = float(config['post-analysis/reanalysis']['signal_to_noise_threshold'])
@@ -402,7 +439,7 @@ def config_handler(from_GUI = False, param_file_path = ''):
             for i_i in range(len(metaboanalyst_groups)-1, -1, -1):
                 metaboanalyst_groups[i_i] = metaboanalyst_groups[i_i].strip()
                 metaboanalyst_groups[i_i] = metaboanalyst_groups[i_i].strip("'")
-                metaboanalyst_groups[i_i] = metaboanalyst_groups[i_i].strip("\"")
+                metaboanalyst_groups[i_i] = metaboanalyst_groups[i_i].strip('"')
                 if len(metaboanalyst_groups[i_i]) == 0:
                     del metaboanalyst_groups[i_i]
             plot_metaboanalyst = (plot_metaboanalyst_file, metaboanalyst_groups)
@@ -410,12 +447,12 @@ def config_handler(from_GUI = False, param_file_path = ''):
         output_plot_data = config['post-analysis/reanalysis'].getboolean('output_plot_data')
         
     if from_GUI:
-        return (custom_glycans_list, min_max_monos, min_max_hex, min_max_hexnac, min_max_sia, min_max_fuc, min_max_ac, min_max_gc, force_nglycan, max_adducts, adducts_exclusion, max_charges, reducing_end_tag, permethylated, reduced, lactonized_ethyl_esterified, fast_iso, high_res, internal_standard, imp_exp_library, exp_lib_name, library_path, only_gen_lib, min_max_xyl), (multithreaded_analysis, number_cores, analyze_ms2, reporter_ions, tolerance, ret_time_interval, rt_tolerance_frag, min_isotopologue_peaks, min_ppp, close_peaks, align_chromatograms, percentage_auc, max_ppm, iso_fit_score, curve_fit_score, s_to_n, custom_noise, samples_path, save_path, plot_metaboanalyst, compositions, iso_fittings, reanalysis, reanalysis_path, output_plot_data)
+        return (custom_glycans_list, min_max_monos, min_max_hex, min_max_hexnac, min_max_sia, min_max_fuc, min_max_ac, min_max_gc, forced, max_adducts, adducts_exclusion, max_charges, reducing_end_tag, permethylated, reduced, lactonized_ethyl_esterified, fast_iso, high_res, internal_standard, imp_exp_library, exp_lib_name, library_path, only_gen_lib, min_max_xyl, min_max_hn, min_max_ua, min_max_sulfation, min_max_phosphorylation), (multithreaded_analysis, number_cores, analyze_ms2, reporter_ions, tolerance, ret_time_interval, rt_tolerance_frag, min_isotopologue_peaks, min_ppp, close_peaks, align_chromatograms, percentage_auc, max_ppm, iso_fit_score, curve_fit_score, s_to_n, custom_noise, samples_path, save_path, plot_metaboanalyst, compositions, iso_fittings, reanalysis, reanalysis_path, output_plot_data)
                 
     #args to execution functions:
-    output_filtered_data_args = [curve_fit_score, iso_fit_score, s_to_n, max_ppm, percentage_auc, reanalysis, reanalysis_path, save_path, analyze_ms2[0], analyze_ms2[2], reporter_ions, plot_metaboanalyst, compositions, align_chromatograms, force_nglycan, ret_time_interval[2], rt_tolerance_frag, iso_fittings, output_plot_data, multithreaded_analysis, number_cores, 0.0]
+    output_filtered_data_args = [curve_fit_score, iso_fit_score, s_to_n, max_ppm, percentage_auc, reanalysis, reanalysis_path, save_path, analyze_ms2[0], analyze_ms2[2], reporter_ions, plot_metaboanalyst, compositions, align_chromatograms, forced, ret_time_interval[2], rt_tolerance_frag, iso_fittings, output_plot_data, multithreaded_analysis, number_cores, 0.0, min_samples, None]
 
-    imp_exp_gen_library_args = [custom_glycans_list, min_max_monos, min_max_hex, min_max_hexnac, min_max_xyl, min_max_sia, min_max_fuc, min_max_ac, min_max_gc, force_nglycan, max_adducts, adducts_exclusion, max_charges, reducing_end_tag, fast_iso, high_res, imp_exp_library, library_path, exp_lib_name, only_gen_lib, save_path, internal_standard, permethylated, lactonized_ethyl_esterified, reduced]
+    imp_exp_gen_library_args = [custom_glycans_list, min_max_monos, min_max_hex, min_max_hexnac, min_max_xyl, min_max_sia, min_max_fuc, min_max_ac, min_max_gc, min_max_hn, min_max_ua, forced, max_adducts, adducts_exclusion, max_charges, reducing_end_tag, fast_iso, high_res, imp_exp_library, library_path, exp_lib_name, only_gen_lib, save_path, internal_standard, permethylated, lactonized_ethyl_esterified, reduced, min_max_sulfation, min_max_phosphorylation, None]
 
     list_of_data_args = [samples_list]
 
@@ -423,10 +460,10 @@ def config_handler(from_GUI = False, param_file_path = ''):
 
     index_spectra_from_file_ms2_args = [None, 2, multithreaded_analysis, number_cores]
 
-    analyze_files_args = [None, None, None, None, tolerance, ret_time_interval, min_isotopologue_peaks, min_ppp, max_charges, custom_noise, close_peaks, multithreaded_analysis, number_cores]
+    analyze_files_args = [None, None, None, None, tolerance, ret_time_interval, min_isotopologue_peaks, min_ppp, max_charges, custom_noise, close_peaks, multithreaded_analysis, number_cores, None, None]
 
-    analyze_ms2_args = [None, None, None, ret_time_interval, tolerance, min_max_monos, min_max_hex, min_max_hexnac, min_max_xyl,  min_max_sia, min_max_fuc, min_max_ac, min_max_gc, max_charges, reducing_end_tag, force_nglycan, permethylated, reduced, lactonized_ethyl_esterified, analyze_ms2[1], analyze_ms2[2], ret_time_interval[2], multithreaded_analysis, number_cores]
+    analyze_ms2_args = [None, None, None, ret_time_interval, tolerance, min_max_monos, min_max_hex, min_max_hexnac, min_max_xyl,  min_max_sia, min_max_fuc, min_max_ac, min_max_gc, min_max_hn, min_max_ua, max_charges, reducing_end_tag, forced, permethylated, reduced, lactonized_ethyl_esterified, analyze_ms2[1], analyze_ms2[2], ret_time_interval[2], multithreaded_analysis, number_cores, None, None]
 
-    arrange_raw_data_args = [None, samples_names, analyze_ms2[0], save_path, [(custom_glycans_list, min_max_monos, min_max_hex, min_max_hexnac, min_max_sia, min_max_fuc, min_max_ac, min_max_gc, force_nglycan, max_adducts, adducts_exclusion, max_charges, reducing_end_tag, permethylated, reduced, lactonized_ethyl_esterified, fast_iso, high_res, internal_standard, imp_exp_library, exp_lib_name, library_path, only_gen_lib, min_max_xyl), (multithreaded_analysis, number_cores, analyze_ms2, reporter_ions, tolerance, ret_time_interval, rt_tolerance_frag, min_isotopologue_peaks, min_ppp, close_peaks, align_chromatograms, percentage_auc, max_ppm, iso_fit_score, curve_fit_score, s_to_n, custom_noise, samples_path, save_path, plot_metaboanalyst, compositions, iso_fittings, reanalysis, reanalysis_path, output_plot_data)]]
+    arrange_raw_data_args = [None, samples_names, analyze_ms2[0], save_path, [[custom_glycans_list, min_max_monos, min_max_hex, min_max_hexnac, min_max_sia, min_max_fuc, min_max_ac, min_max_gc, forced, max_adducts, adducts_exclusion, max_charges, reducing_end_tag, permethylated, reduced, lactonized_ethyl_esterified, fast_iso, high_res, internal_standard, imp_exp_library, exp_lib_name, library_path, only_gen_lib, min_max_xyl, min_max_hn, min_max_ua, min_max_sulfation, min_max_phosphorylation], [multithreaded_analysis, number_cores, analyze_ms2, reporter_ions, tolerance, ret_time_interval, rt_tolerance_frag, min_isotopologue_peaks, min_ppp, close_peaks, align_chromatograms, percentage_auc, max_ppm, iso_fit_score, curve_fit_score, s_to_n, custom_noise, samples_path, save_path, plot_metaboanalyst, compositions, iso_fittings, reanalysis, reanalysis_path, output_plot_data]], None, None]
 
     return output_filtered_data_args, imp_exp_gen_library_args, list_of_data_args, index_spectra_from_file_ms1_args, index_spectra_from_file_ms2_args, analyze_files_args, analyze_ms2_args, arrange_raw_data_args, samples_names, reanalysis, analyze_ms2[0]
