@@ -44,7 +44,7 @@ class make_mzxml(object):
     
     Parameters
     ----------
-    object : string
+    path_to_mzml : string
         A string containing the path to the mzML file.
         
     Uses
@@ -145,7 +145,7 @@ def eic_from_glycan(files,
                     threads_arrays,
                     rt_arrays,
                     ms1_id,
-                    rewind = 10):
+                    sampling_rates):
     '''Generates a very processed EIC for each adduct of each glycan of each sample.
     Removes non-monoisotopic peaks, check charges and calculates multiple quality scoring
     data.
@@ -154,6 +154,9 @@ def eic_from_glycan(files,
     ----------
     files : list
         A list of generators obtained from the function pyteomics.mzxml.MzXML().
+        
+    glycan : str
+        The name of the target glycan.
 
     glycan_info : dict
         A dictionary containing multiple glycan info, including the needed 'Adducts_mz'
@@ -197,6 +200,9 @@ def eic_from_glycan(files,
     
     ms1_id : list
         List of the MS1 spectra that will be analyzed, synchronized to the threads_array.
+        
+    rewind : int
+        The maximum number of spectras it can go back to for retroactively checking after finding a peak.
 
     Uses
     ----
@@ -237,6 +243,9 @@ def eic_from_glycan(files,
         dict with keys for each file id, at which each value corresponds to a given 
         calculated isotopic fitting score for the retention time at the same index.
         
+    verbose_info : list
+        Legacy list used to save verbose info to. Unused and in need of CLEANUP.
+        
     raw_data : dict
         A dictionary with keys for each adduct combo of a given glycan, which value is a
         dict with keys for each file id, at which value is a list of lists, with the
@@ -293,34 +302,47 @@ def eic_from_glycan(files,
                                  ms1_id[j_j][k_k],
                                  ms1_id[-1],
                                  adduct_mass,
-                                 adduct_charge)
-                if len(buffer) > 4 and buffer[-1] != None and buffer[-2] != None and buffer[-3] != None and buffer[-4] == None:
-                    for l_l in range(-3, -3-rewind, -1):
-                        if l_l == -len(buffer) or buffer[l_l-1] != None: # the l_l limit is set to -10 to be compatible with buffer_size
+                                 adduct_charge,
+                                 sampling_rates)
+                if len(buffer) > max([4*sampling_rates[j_j], 4]):
+                    no_rewind = False
+                    for l in range(-1, -(4*sampling_rates[j_j])-2, -1):
+                        if buffer[l] == None and l != -(max([4*sampling_rates[j_j], 4]))-1:
+                            no_rewind = True
                             break
-                        analyze_mz_array(j[thread_numbers[k_k+l_l]]['m/z array'],
-                                         j[thread_numbers[k_k+l_l]]['intensity array'],
-                                         glycan_info,
-                                         tolerance,
-                                         min_isotops,
-                                         noise,
-                                         avg_noise,
-                                         max_charges,
-                                         ppm_info,
-                                         iso_fitting_quality,
-                                         data,
-                                         raw_data,
-                                         isotopic_fits,
-                                         i,
-                                         j_j,
-                                         thread_numbers[k_k+l_l],
-                                         j[thread_numbers[k_k+l_l]]['retentionTime'],
-                                         ms1_id[j_j][k_k+l_l],
-                                         ms1_id[-1],
-                                         adduct_mass,
-                                         adduct_charge,
-                                         retest = True,
-                                         retest_no = l_l-1)
+                        if buffer[l] != None and l == -(max([4*sampling_rates[j_j], 4]))-1:
+                            no_rewind = True
+                            break
+                    if not no_rewind:
+                        for l_l in range(-max([4*sampling_rates[j_j], 4]), -max([4*sampling_rates[j_j], 4])-len(buffer), -1):
+                            if l_l == -len(buffer) or buffer[l_l-1] != None:
+                                break
+                            analyze_mz_array(j[thread_numbers[k_k+l_l]]['m/z array'],
+                                             j[thread_numbers[k_k+l_l]]['intensity array'],
+                                             glycan_info,
+                                             tolerance,
+                                             min_isotops,
+                                             noise,
+                                             avg_noise,
+                                             max_charges,
+                                             ppm_info,
+                                             iso_fitting_quality,
+                                             data,
+                                             raw_data,
+                                             isotopic_fits,
+                                             i,
+                                             j_j,
+                                             thread_numbers[k_k+l_l],
+                                             j[thread_numbers[k_k+l_l]]['retentionTime'],
+                                             ms1_id[j_j][k_k+l_l],
+                                             ms1_id[-1],
+                                             adduct_mass,
+                                             adduct_charge,
+                                             sampling_rates,
+                                             retest = True,
+                                             retest_no = l_l-1)
+                            if buffer[l_l-1] == None:
+                                break
     return data, ppm_info, iso_fitting_quality, verbose_info, raw_data, isotopic_fits
 
     
@@ -345,6 +367,7 @@ def analyze_mz_array(sliced_mz,
                      last_ms1_id,
                      adduct_mass,
                      adduct_charge,
+                     sampling_rates,
                      filtered = True,
                      retest = False,
                      retest_no = 0):
@@ -425,6 +448,15 @@ def analyze_mz_array(sliced_mz,
         
     adduct_charge : int
         The charge of the adduct.
+        
+    filtered : boolean
+        Whether the output will be filtered or not. Non-filtered output allows for MUCH FASTER tracing. Might get used in the future.
+        
+    retest : boolean
+        Whether this is a retest check or not. For use with rewind on eic_from_glycan.
+        
+    retest_no : int
+        Buffer ID of the retest.
     
     Uses
     ----
@@ -482,7 +514,7 @@ def analyze_mz_array(sliced_mz,
                 
             if not bad:
                 for i in charge_range: #check if it's monoisotopic and correct charge
-                    if (len(buffer) <= 2 or (len(buffer) > 2 and buffer[-3] == None)) and not retest: #only check if it's monoisotopic if at least one of the last 3 RT got nothing...
+                    if (len(buffer) <= round(max([2*sampling_rates[file_id], 2])) or (len(buffer) > round(max([2*sampling_rates[file_id], 2])) and buffer[round(-max([2*sampling_rates[file_id], 2]))-1] == None)) and not retest: #only check if it's monoisotopic if at least one of the last 3 RT got nothing...
                         temp_id = General_Functions.binary_search_with_tolerance(sliced_mz, found_mz-(General_Functions.h_mass/i), 0, mz_id, General_Functions.tolerance_calc(tolerance[0], tolerance[1], found_mz-(General_Functions.h_mass/i)), sliced_int) #check monoisotopic
                         if temp_id != -1 and sliced_int[temp_id] > 0:
                             expected_value = (sliced_mz[temp_id]*i*0.0006)+0.1401 #based on linear regression of the relationship between masses and the second isotopic peak relative intensity of the average of different organic macromolecules
@@ -509,7 +541,7 @@ def analyze_mz_array(sliced_mz,
                 for i_i, i in enumerate(glycan_info['Isotopic_Distribution_Masses']): #check isotopic peaks and add to the intensity
                     if i_i == 0: #ignores monoisotopic this time around
                         continue
-                    temp_id = General_Functions.binary_search_with_tolerance(sliced_mz, found_mz+(i_i*(General_Functions.h_mass/abs(adduct_charge))), mz_id, sliced_mz_length, General_Functions.tolerance_calc(tolerance[0], tolerance[1], found_mz+(i_i*(General_Functions.h_mass/abs(adduct_charge)))), sliced_int)
+                    temp_id = General_Functions.binary_search_with_tolerance(sliced_mz, found_mz+(i_i*(General_Functions.h_mass/abs(adduct_charge))), mz_id, sliced_mz_length, General_Functions.tolerance_calc(tolerance[0], tolerance[1], found_mz+(i_i*(General_Functions.h_mass/abs(adduct_charge)))), sliced_int, mz_isos)
                     if temp_id != -1 and sliced_int[temp_id] > 0:
                         isos_found += 1
                         mz_isos.append(sliced_mz[temp_id])
@@ -527,7 +559,7 @@ def analyze_mz_array(sliced_mz,
                             
             if not bad and (iso_actual[1] < 0.2 or iso_actual[1] > 5): #this should avoid situations where it's obvious that it's picking the wrong charge because the second peak is almost invisible compared to the third and first, which when z=2 means that it's very likely actually a singly charge compound, for example
                 if len(iso_actual) > 2:
-                    if iso_actual[2] > iso_actual[1]*10 or iso_actual[1] > 5:
+                    if iso_actual[2] > iso_actual[1]*10 or iso_actual[1] < 0.2 or iso_actual[1] > 5:
                         bad = True
                 else:
                     if iso_actual[1] < 0.2 or iso_actual[1] > 5: #smallest glycan should have the second iso_actual somewhere around 0.5, so a cutoff lower than that is fine
@@ -560,7 +592,7 @@ def analyze_mz_array(sliced_mz,
                 if len(iso_actual) == 2:
                     iso_quali = (iso_quali*0.8)
                 
-                mz_isos = [glycan_info['Adducts_mz'][glycan_id]]+mz_isos
+                mz_isos = [found_mz]+mz_isos
                 
                 if not retest:
                     buffer.append(([glycan_id, file_id, ms1_id, float("%.4f" % round(ret_time, 4))], [ppm_error, iso_quali, intensity, [mz_isos, iso_target, iso_actual, iso_quali]]))
@@ -578,8 +610,8 @@ def analyze_mz_array(sliced_mz,
     # print(f"Buffer before clean-up: {buffer}")
     
     #dynamical clean-up of buffer
-    min_in_a_row = 4
-    buffer_size = 10
+    min_in_a_row = round(max([4*sampling_rates[file_id], 4]))
+    buffer_size = round(max([10*sampling_rates[file_id], 10]))
     if filtered:
         in_a_row = 0
         for i_i, i in enumerate(buffer):
@@ -684,6 +716,7 @@ def peak_curve_fit(rt_int,
     '''
     x = rt_int[0][peak['peak_interval_id'][0]:peak['peak_interval_id'][1]+1]
     y = rt_int[1][peak['peak_interval_id'][0]:peak['peak_interval_id'][1]+1]
+    baseline_correction = min(y[0], y[-1])
     interval = x[len(x)//2]-x[(len(x)//2)-1]
     fits_list = []
     for m in range(1, 11):
@@ -698,16 +731,17 @@ def peak_curve_fit(rt_int,
             before = sorted(before)
             temp_x = before+x+after
             temp_y = y_adds+y+y_adds
-            max_amp_id = temp_y.index(max(temp_y))
-            counts_max = 0
-            for i in temp_y:
-                if i == temp_y[max_amp_id]:
-                    counts_max += 1
-            max_amp_id += int(counts_max/2)
+            max_amp = max(temp_y)
+            maximums = []
+            for index, i in enumerate(temp_y):
+                if i > max_amp*0.8:
+                    maximums.append(index)
+            max_amp_id = round(sum(maximums)/len(maximums))
             y_gaussian = []
             for i in temp_x:
                 y_gaussian.append(General_Functions.normpdf(i, temp_x[max_amp_id], (temp_x[-1]-temp_x[0])/m))
-            scaler = (temp_y[max_amp_id]/y_gaussian[max_amp_id])
+            y_gaussian = [i-min(y_gaussian) for i in y_gaussian]
+            scaler = (max_amp/y_gaussian[max_amp_id])
             y_gaussian_scaled = []
             for j in y_gaussian:
                 y_gaussian_scaled.append(j*scaler)
@@ -744,6 +778,9 @@ def iso_fit_score_calc(iso_fits,
     ----
     numpy.average : float
         Calculates the weighted average of an array based on another array of weights.
+    
+    weights_list : list
+        A list containing the weights used to calculate the scores. Comes from the curve fitting.
         
     Returns
     -------
@@ -772,6 +809,9 @@ def average_ppm_calc(ppm_array,
         
     peak : dict
         A dictionary containing all sorts of identified peaks information.
+    
+    weights_list : list
+        A list containing the weights used to calculate the scores. Comes from the curve fitting.
 
     Uses
     ----
@@ -815,12 +855,16 @@ def peaks_from_eic(rt_int,
     ----------
     rt_int : list
         A list containing two lists: the first one has all the retention times, the
-        second one has all the intensities.
+        second one has all the FILTERD BUT UNSMOOTHED intensities.
         
     rt_int_smoothed : list
         A list containing two lists: the first one has all the retention times, the
         second one has all the SMOOTHED intensities.
         
+    raw_rt_int : list
+        A list containing two lists: the first one has all the retention times, the
+        second one has all the UNFILTERED intensities. Used to calculate signal-to-noise ratio.
+        
     rt_interval : tuple
         A tuple where the first index contains the beggining time of the retention time
         interval you wish to analyze and the second contains the end time.
@@ -924,7 +968,7 @@ def peaks_from_eic(rt_int,
     for i_i, i in enumerate(peaks_ranges):
         if i_i % 2 == 0:
             peak_limits = [i, peaks_ranges[i_i+1]]
-            if peak_limits[0] == peak_limits[1]:
+            if peak_limits[0] == peak_limits[1] or (min_ppp[0] and peak_limits[1] - peak_limits[0] < min_ppp[1]):
                 continue
             temp_peak_width = (rt_int[0][peak_limits[1]]-rt_int[0][peak_limits[0]])
             peaks.append({'id': i, 'rt': rt_int[0][rt_int_smoothed[1].index(max(rt_int_smoothed[1][peak_limits[0]:peak_limits[1]+1]))], 'int': max(raw_rt_int[1][peak_limits[0]:peak_limits[1]+1]), 'peak_width': temp_peak_width, 'peak_interval': (rt_int[0][peak_limits[0]], rt_int[0][peak_limits[1]]), 'peak_interval_id': (peak_limits[0], peak_limits[1])})
