@@ -25,34 +25,42 @@ from statistics import stdev, mean
 from scipy.stats import linregress
 from scipy.sparse.linalg import splu
 from scipy import sparse
+import dill
 import numpy
 import sys
 import datetime
 import xlsxwriter
 import zipfile
 import pathlib
+import tempfile
+import shutil
 import os
+import copy
 
 ##---------------------------------------------------------------------------------------
 ##Hard-coded permanent information
 
 monosaccharides = {
-    "H": ("Hexose", "C6O6H12", {"C": 6, "O": 5, "N": 0, "H": 10}),
-    "N": ("N-Acetyl Hexosamine", "C8O6NH15", {"C": 8, "O": 5, "N": 1, "H": 13}),
-    "X": ("Xylose", "C5O5H10", {"C": 5, "O": 4, "N": 0, "H": 8}),
-    "S": ("Acetyl Neuraminic Acid", "C11O9NH19", {"C": 11, "O": 8, "N": 1, "H": 17}),
-    "Am": ("Lactonized Acetyl Neuraminic Acid alpha2,3 bound", "C11O8N2H20", {"C": 11, "O": 7, "N": 2, "H": 18}),
-    "E": ("Ethyl-Esterified Acetyl Neuraminic Acid alpha2,6 bound", "C13O9NH23", {"C": 13, "O": 8, "N": 1, "H": 21}),
-    "F": ("Fucose", "C6O5H12", {"C": 6, "O": 4, "N": 0, "H": 10}),
-    "G": ("Glycolyl Neuraminic Acid", "C11O10NH19", {"C": 11, "O": 9, "N": 1, "H": 17}),
-    "AmG": ("Lactonized Glycolyl Neuraminic Acid alpha2,3 bound", "C11O8N2H20", {"C": 11, "O": 8, "N": 2, "H": 18}),
-    "EG": ("Ethyl-Esterified Glycolyl Neuraminic Acid alpha2,6 bound", "C13O9NH23", {"C": 13, "O": 9, "N": 1, "H": 21})
+    "H": ("Hexose", "C6O6H12", {"C": 6, "O": 5, "N": 0, "H": 10}, "H"),
+    "N": ("N-Acetyl Hexosamine", "C8O6NH15", {"C": 8, "O": 5, "N": 1, "H": 13}, "N"),
+    "X": ("Xylose", "C5O5H10", {"C": 5, "O": 4, "N": 0, "H": 8}, "X"),
+    "S": ("Acetyl Neuraminic Acid", "C11O9NH19", {"C": 11, "O": 8, "N": 1, "H": 17}, "S"),
+    "Am": ("Lactonized Acetyl Neuraminic Acid alpha2,3 bound", "C11O8N2H20", {"C": 11, "O": 7, "N": 2, "H": 18}, "L"),
+    "E": ("Ethyl-Esterified Acetyl Neuraminic Acid alpha2,6 bound", "C13O9NH23", {"C": 13, "O": 8, "N": 1, "H": 21}, "E"),
+    "F": ("Deoxyhexose", "C6O5H12", {"C": 6, "O": 4, "N": 0, "H": 10}, "F"),
+    "G": ("Glycolyl Neuraminic Acid", "C11O10NH19", {"C": 11, "O": 9, "N": 1, "H": 17}, "G"),
+    "AmG": ("Lactonized Glycolyl Neuraminic Acid alpha2,3 bound", "C11O8N2H20", {"C": 11, "O": 8, "N": 2, "H": 18}, "A"),
+    "EG": ("Ethyl-Esterified Glycolyl Neuraminic Acid alpha2,6 bound", "C13O9NH23", {"C": 13, "O": 9, "N": 1, "H": 21}, "R"),
+    "HN": ("Hexosamine", "C6O5NH13", {"C": 6, "O": 4, "N": 1, "H": 11}, "M"),
+    "UA": ("Uronic Acid", "C6O7H10", {"C": 6, "O": 6, "N": 0, "H": 8}, "U")
     }
 '''A hardcoded dictionary containing each single letter code for monosaccharides as key
 and a tuple containing the full monosaccharide name, its full molecular formula and its
 residue composition in dict form as value.
 '''
 
+default_composition = {"H": 0, "N": 0, "X": 0, "S": 0, "Am": 0, "E": 0, "F": 0, "G": 0, "AmG": 0, "EG": 0, "HN": 0, "UA": 0}
+
 h_mass = mass.calculate_mass(composition={'H' : 1})
 '''The mass of an hydrogen-1 atom. Pre-calculated here to avoid calculating too many
 times during a run.
@@ -62,8 +70,41 @@ times during a run.
 ##General functions (these functions use only external libraries, such as itertools and
 ##pyteomics).
 
-def binary_search_with_tolerance(arr, target, low, high, tolerance, int_arr = []):
-    '''
+def binary_search_with_tolerance(arr, target, low, high, tolerance, int_arr = [], black_list = []):
+    '''A function to quickly find a target in an array by splitting the array recurssively in two and looking for the mid point, finding out if value is bigger or smaller than  target, then splitting again. It also checks if found target in target array is within a tolerance, and picks the most intense one within the tolerance if intensity array is available, else picks the closest one to target.
+    
+    Parameters
+    ----------
+    arr : list/np.array
+        Target array to search for target.
+        
+    target : float
+        Float to find in target array.
+        
+    low : int
+        Index of the first element in target array.
+        
+    high : int
+        Index of the last element in target array.
+        
+    tolerance : float
+        Tolerance to check for target in target array.
+        
+    int_arr : list/np.array
+        List of intensities, synchronized with arr, to help choose the best target within tolerance.
+        
+    Uses
+    ----
+    numpy.argmax : int
+        Outputs the index of the highest value in a given array.
+        
+    numpy.argmin : int
+        Outputs the index of the value with smallest difference to a given target, in an array.
+        
+    Returns
+    -------
+    selected_id : index
+        The index of the selected target.
     '''
     # Base case: if the range is invalid, the target is not in the array
     if low > high:
@@ -74,6 +115,7 @@ def binary_search_with_tolerance(arr, target, low, high, tolerance, int_arr = []
     
     # Check if the target is within the tolerance range of the middle element
     if abs(arr[mid] - target) <= tolerance:
+    
         range_width = 5
         range_search = [mid, mid+1]
         for i in range(mid, mid-range_width, -1):
@@ -84,24 +126,44 @@ def binary_search_with_tolerance(arr, target, low, high, tolerance, int_arr = []
             range_search[1] = i
             if i > high or arr[i] > target+tolerance:
                 break
+                
         if len(int_arr) != 0:
             array_slice = int_arr[range_search[0]:range_search[1]]
         else:
             array_slice = arr[range_search[0]:range_search[1]]
+            
         if len(array_slice) == 0:
             return -1
+
         if len(int_arr) != 0:
-            selected_id = range_search[0]+numpy.argmax(array_slice)
+            relative_id = numpy.argmax(array_slice)
         else:
-            selected_id = range_search[0]+(numpy.abs(array_slice - target).argmin())
-        
+            relative_id = (numpy.abs(array_slice - target).argmin())
+        selected_id = range_search[0]+relative_id
+            
+        # This avoids picking the same peak twice
+        forbidden_ids = []
+        while arr[selected_id] in black_list:
+            forbidden_ids.append(relative_id)
+            if len(int_arr) != 0:
+                array_slice[relative_id] = 0
+            else:
+                array_slice[relative_id] += 1000
+            if len(int_arr) != 0:
+                relative_id = numpy.argmax(array_slice)
+            else:
+                relative_id = (numpy.abs(array_slice - target).argmin())
+            if relative_id in forbidden_ids:
+                return -1
+            selected_id = range_search[0]+relative_id
+            
         return selected_id
     elif arr[mid] < target:
         # If target is greater, ignore the left half
-        return binary_search_with_tolerance(arr, target, mid + 1, high, tolerance, int_arr)
+        return binary_search_with_tolerance(arr, target, mid + 1, high, tolerance, int_arr, black_list)
     else:
         # If target is smaller, ignore the right half
-        return binary_search_with_tolerance(arr, target, low, mid - 1, tolerance, int_arr)
+        return binary_search_with_tolerance(arr, target, low, mid - 1, tolerance, int_arr, black_list)
 
 def linear_regression(x, y, th = 2.5):
     '''Traces a linear regression of supplied 2d data points and returns the slope,
@@ -137,7 +199,7 @@ def linear_regression(x, y, th = 2.5):
         y-intercept of the fitted linear equation curve.
         
     outlier_indices : list
-        A list of the residuals outside the given threshold of the linear equation.
+        An index list of the residuals outside the given threshold of the linear equation.
     '''
     if len(x) != len(y): # Ensure x and y have the same length
         raise ValueError("Input arrays x and y must have the same length.")
@@ -156,7 +218,34 @@ def linear_regression(x, y, th = 2.5):
     return m, b, outlier_indices
     
 def cleanup_by_average(x, y, th = 2.5):
-    '''
+    '''Used to find the outliers of a given fit average.
+    
+    Parameters
+    ----------
+    x : list
+        A list containing the x elements of the datapoints.
+        
+    y : list
+        A list containing the y elements of the datapoints.
+        
+    th : float
+        The threshold for calculating outliers.
+    
+    Uses
+    ----
+    numpy.array : ndarray
+        Transforms a list into a numpy array.
+        
+    Returns
+    -------
+    mean : float
+        The mean of the fit.
+        
+    residuals : float
+        Values of residuals.
+        
+    outlier_indices : list
+        An index list of the residuals outside the given threshold.
     '''
     if len(x) != len(y): # Ensure x and y have the same length
         raise ValueError("Input arrays x and y must have the same length.")
@@ -173,7 +262,7 @@ def cleanup_by_average(x, y, th = 2.5):
     # Final equation is used as (y = mx + b)
     return mean, residuals, outlier_indices
     
-def make_gg(temp_dir, save_dir, filename):
+def make_gg(temp_dir, save_dir, filename, ignore_files = False):
     '''Creates a zipped file with extension .gg containing raw_data files.
     
     Parameters
@@ -201,11 +290,15 @@ def make_gg(temp_dir, save_dir, filename):
         Creates filename.gg file, containing files in temp_dir, at the save_dir.
     '''
     files_list = os.listdir(temp_dir)
-    with zipfile.ZipFile(save_dir+filename+".gg", 'w', compression=zipfile.ZIP_DEFLATED) as zipf:
+    with zipfile.ZipFile(os.path.join(save_dir, filename+".gg"), 'w', compression=zipfile.ZIP_DEFLATED) as zipf:
         for file in files_list:
-            zipf.write(temp_dir+"/"+str(file), arcname=file)
+            if ignore_files:
+                if file != "metadata" and file != "results" and "_" not in file:
+                    continue
+            zipf.write(os.path.join(temp_dir, file), arcname=file)
+        zipf.close()
             
-def open_gg(gg_file, temp_path):
+def open_gg(gg_file, temp_path, file = 'all'):
     '''Unzipts the .gg file to temp_path.
     
     Parameters
@@ -230,8 +323,23 @@ def open_gg(gg_file, temp_path):
         Extracts raw_data files from gg_file to temp_path.
     '''
     with zipfile.ZipFile(gg_file, 'r') as zip_ref:
-        zip_ref.extractall(temp_path)
-    
+        if file == 'all':
+            zip_ref.extractall(temp_path)
+        else:
+            zip_ref.extract(file, temp_path)
+        zip_ref.close()
+        
+def access_chromatogram(file_number, chromatogram_name, temp_folder, reanalysis_path):
+    '''
+    '''
+    if chromatogram_name not in os.listdir(temp_folder):
+        if f'{file_number}_eics' not in os.listdir(temp_folder):
+            open_gg(reanalysis_path, temp_folder, f'{file_number}_eics')
+        open_gg(os.path.join(temp_folder, f'{file_number}_eics'), temp_folder, chromatogram_name)
+    with open(os.path.join(temp_folder, chromatogram_name), 'rb') as f:
+        chromatogram = dill.load(f)
+        f.close()
+    return chromatogram
 
 def calculate_ppm_diff(mz, target):
     '''Calculates the PPM difference between a mz and a target mz.
@@ -473,7 +581,7 @@ def normpdf(x, mean, sd):
     num = exp(-(float(x)-float(mean))**2/(2*var))
     return num/denom
 
-def form_to_comp(string):
+def form_to_comp(string, form_type = 'glycan'):
     '''Separates a molecular formula or monosaccharides formula of glycans into a
     dictionary with each atom/monosaccharide as a key and its amount as value.
 
@@ -488,6 +596,7 @@ def form_to_comp(string):
         A dictionary with keys containing the monosaccharides/atoms letter(s) and values
         containing the amounts of each. ie. {"H": 5, "N": 4, "S": 1, "F": 1, "G": 1}.
     '''
+    string = string.split("+")[0] #avoids getting phosphorylation and sulfation symbols
     counts = {}
     split_str = split('(\\d+)', string)
     negative = False
@@ -558,25 +667,23 @@ def glycan_to_atoms(glycan_composition, permethylated):
         "O": 6, "N": 0, "H": 12}.
     '''
     atoms = {"C": 0, "O": 0, "N": 0, "H": 0}
+    monosaccharides_local = copy.deepcopy(monosaccharides)
     if permethylated:
-        for i in monosaccharides:
-            if i == 'H':
-                monosaccharides[i][2]['C'] += 3
-                monosaccharides[i][2]['H'] += 6
-            if i == 'N':
-                monosaccharides[i][2]['C'] += 3
-                monosaccharides[i][2]['H'] += 6
+        for i in monosaccharides_local:
+            if i == 'H' or i == 'N':
+                monosaccharides_local[i][2]['C'] += 3
+                monosaccharides_local[i][2]['H'] += 6
             if i == 'F':
-                monosaccharides[i][2]['C'] += 2
-                monosaccharides[i][2]['H'] += 4
+                monosaccharides_local[i][2]['C'] += 2
+                monosaccharides_local[i][2]['H'] += 4
             if i == 'S' or i == 'G':
-                monosaccharides[i][2]['C'] += 5
-                monosaccharides[i][2]['H'] += 10
+                monosaccharides_local[i][2]['C'] += 5
+                monosaccharides_local[i][2]['H'] += 10
     for i in glycan_composition:
         if i == "T":
             continue
         for j in atoms:
-            atoms[j] += monosaccharides[i][2][j]*glycan_composition[i]
+            atoms[j] += monosaccharides_local[i][2][j]*glycan_composition[i]
     return atoms
 
 def count_seq_letters(string):
@@ -648,7 +755,7 @@ def sum_monos(*compositions):
     summed_comp : dict
         Dictionary containing the sum of each monosaccharides of the compositions.
     '''
-    summed_comp = {"H": 0, "N": 0, "X": 0, "S": 0, "Am": 0, "E": 0, "F": 0, "G": 0, "AmG": 0, "EG": 0, "T": 0}
+    summed_comp = {"H": 0, "N": 0, "X": 0, "S": 0, "Am": 0, "E": 0, "F": 0, "G": 0, "AmG": 0, "EG": 0, "T": 0, "HN": 0, "UA": 0}
     for i in compositions:
         for j in i:
             summed_comp[j]+=i[j]
@@ -699,16 +806,41 @@ def calculate_comp_from_mass(tag_mass):
         to the tag's added mass in dictionary form and the calculated hypothetical mass
         of the molecule as float.
     '''
-    closest = ({}, 0)
-    test_tag_mass = 0
-    atoms_number = int(tag_mass/5)
-    while atoms_number > atoms_number/2:
-        for i in combinations_with_replacement("CONH", atoms_number):
-            seq_readable = count_seq_letters("".join(i))
-            test_tag_mass = mass.calculate_mass(composition = seq_readable)
-            if abs(test_tag_mass-tag_mass) < abs(closest[1]-tag_mass):
-                closest = (seq_readable, test_tag_mass)
-        atoms_number -= 1      
+    elements = "CONH"
+    closest = ({}, float('inf'))
+    
+    # Estimate minimum and maximum number of carbon atoms
+    min_carbons = int(((0.05 * tag_mass) + 0.5) * 0.7)
+    max_carbons = int(((0.05 * tag_mass) + 0.5) * 1.3)
+
+    max_nitrogens = max(3, int(tag_mass/200))  # Maximum number of nitrogen atoms allowed
+    min_atoms = max(min_carbons, int(tag_mass/10))  # Min total atoms considered per combination
+    max_atoms = int(tag_mass/5)  # Max total atoms considered per combination
+    
+    # print(f"Carbon range: {min_carbons} to {max_carbons}, Max Nitrogens: {max_nitrogens}, Min atoms: {min_atoms}, Max atoms: {max_atoms}")
+
+    # Generate combinations within the carbon range
+    for c_count in range(min_carbons, max_carbons + 1):
+        for n_count in range(0, max_nitrogens + 1):
+            for total_atoms in range(min_atoms, max_atoms + 1):
+                non_carbon_nitrogen_count = total_atoms - c_count
+                if non_carbon_nitrogen_count < n_count:
+                    continue
+                
+                # Generate combinations of 'O' and 'H', with fixed 'N' count
+                combos = combinations_with_replacement('OH', non_carbon_nitrogen_count - n_count)
+                
+                for combo in combos:
+                    seq_readable = count_seq_letters('C' * c_count + 'N' * n_count + ''.join(combo))
+                    test_tag_mass = mass.calculate_mass(composition=seq_readable)
+                    
+                    if abs(test_tag_mass - tag_mass) < abs(closest[1] - tag_mass):
+                        closest = (seq_readable, test_tag_mass)
+
+                    # Early stopping if we find a very close match
+                    if abs(test_tag_mass - tag_mass) < 1e-3:
+                        return closest
+
     return closest
     
 def calculate_isotopic_pattern(glycan_atoms,
@@ -760,13 +892,16 @@ def calculate_isotopic_pattern(glycan_atoms,
                                           overall_threshold = 1e-4)
     else:
         isotopologue = mass.isotopologues(glycan_atoms, report_abundance = True,
-                                          elements_with_isotopes = ["C", "N", "O", "H"],
+                                          elements_with_isotopes = ["C", "N", "O", "H", "S", "P", "Na"],
                                           overall_threshold = 1e-4)
     isotop_arranged = []
     relative_isotop_pattern = []
     relative_isotop_mass = []
-    for i in isotopologue:
-        isotop_arranged.append({'mz' : mass.calculate_mass(i[0]), 'Ab' : i[1]})
+    try:
+        for i in isotopologue:
+            isotop_arranged.append({'mz' : mass.calculate_mass(i[0]), 'Ab' : i[1]})
+    except Exception:
+        pass
     isotop_arranged = sorted(isotop_arranged, key=lambda x: x['mz'])
     for i_i, i in enumerate(isotop_arranged):
         relative_isotop_pattern.append(i['Ab']/isotop_arranged[0]['Ab'])
