@@ -76,9 +76,11 @@ custom_glycans_list = H3N2, H5N2, H5N4S2F1
 ;
 total_monosaccharides = 5, 22
 hexoses = 3, 10
+hexosamines = 0, 0
 hexnacs = 2, 8
-xylose = 0, 0
+xyloses = 0, 0
 sialic_acids = 0, 4
+uronic_acids = 0, 0
 fucoses = 0, 2
 neu5ac = 0, 4
 neu5gc = 0, 0
@@ -88,12 +90,15 @@ neu5gc = 0, 0
 ;
 
 [common_library_building_settings]
-force_nglycan = yes
-; 	Used to force some monosaccharides compositions
-;	associated with N-Glycans biologically known 
-;	features. If not used, gives a much broader 
-;	library that can be used for analysis of 
-;	O-Glycans, for example.
+force_class_structure = none
+; 	Used to force some monosaccharides compositions 
+;	compatible with structures of N-glycans, O-glycans
+;	or GAGs.
+;	Options:
+;	- none
+;	- n_glycans
+;	- o_glycans
+;	- gags
 ;
 max_adducts = H3
 adducts_exclusion = 
@@ -135,6 +140,14 @@ aminated_ethyl_esterified = no
 ;	be identified as 'Am' and Ethyl Esterified Acetyl
 ;	Sialic Acid will be identified as 'E'.
 ;
+min_max_sulfation_per_glycan = 0, 0
+;	Minimum and maximum amount of sulfation
+;	substituents per glycan.
+;
+min_max_phosphorylation_per_glycan = 0, 0
+;	Minimum and maximum amount of phosphorylation
+;	substituents per glycan.
+;
 fast_iso = yes
 ; 	Allows you to calculate the isotopic distribution
 ;	of glycans based only on carbon isotopes (fast, 
@@ -152,9 +165,13 @@ high_resolution_isotopic_dist = no
 ;	on FT mass spectrometers.
 ;
 internal_standard_mass = 0.0
-; 	If using an internal standard, insert its mass 
-;	here for the script to calculate its area. This 
-;	also allows the script to output a normalized 
+; 	If using an internal standard, insert its mass,
+;	chemical formula or glycan formula here for 
+;	GlycoGenius to calculate its area. 
+;	If a glycan formula is used, reducing end modifications
+;	and permethylation (if selected) are applied to the 
+;	internal standard.
+;	This also allows the script to output a normalized 
 ;	metaboanalyst compatible file.
 ;
 
@@ -186,7 +203,7 @@ accuracy_value = 10
 ;	raw data using max_ppm, but this may lead to 
 ;	false positives.
 ;
-ret_time_interval = 1, 999
+ret_time_interval = 0, 999
 ; 	The minimum and maximum retention time, in MINUTES, 
 ;	used for various portions of the script. A shorter 
 ;	interval of ret_time makes the script run faster, 
@@ -232,6 +249,10 @@ auc_percentage_threshold = 1
 ;	auc_percentage_threshold is set to 1%, every peak 
 ;	with an auc of 1 and below will be supressed)
 ;
+minimum_samples = 0
+;	Filter out glycans that weren't found in at least
+;	a certain number of samples, set by 'minimum_samples'.
+;
 max_ppm_threshold = 10
 ; 	Maximum PPM error for data curation. If value is 
 ;	greater than equivalent accuracy_value, data won't 
